/*************************************************
* MPI Multiply-Add Core Header File              *
* (C) 1999-2006 The Botan Project                *
*************************************************/

#ifndef BOTAN_MP_MADD_H__
#define BOTAN_MP_MADD_H__

#include <botan/mp_types.h>

#if (BOTAN_MP_WORD_BITS != 64)
   #error The mp_asm64 module requires that BOTAN_MP_WORD_BITS == 64
#endif

#if defined(BOTAN_TARGET_ARCH_IS_AMD64)

   #define BOTAN_WORD_MUL(a,b,z1,z0) do {       \
      asm("mulq %3" : "=d" (z0), "=a" (z1) :    \
          "a" (a), "rm" (b) : "cc");            \
   } while(0);

#elif defined(BOTAN_TARGET_ARCH_IS_ALPHA)

   #define BOTAN_WORD_MUL(a,b,z1,z0) do {                   \
      asm("umulh %1,%2,%0" : "=r" (z0) : "r" (a), "r" (b)); \
      z1 = a * b;                                           \
   } while(0);

#elif defined(BOTAN_TARGET_ARCH_IS_IA64)

   #define BOTAN_WORD_MUL(a,b,z1,z0) do {                     \
      asm("xmpy.hu %0=%1,%2" : "=f" (z0) : "f" (a), "f" (b)); \
      z1 = a * b;                                             \
   } while(0);

#elif defined(BOTAN_TARGET_ARCH_IS_PPC64)

   #define BOTAN_WORD_MUL(a,b,z1,z0) do {                           \
      asm("mulhdu %0,%1,%2" : "=r" (z0) : "r" (a), "r" (b) : "cc"); \
      z1 = a * b;                                                   \
   } while(0);

#elif defined(BOTAN_TARGET_ARCH_IS_MIPS64)

   #define BOTAN_WORD_MUL(a,b,z1,z0) do {                            \
      asm("dmultu %2,%3" : "=h" (z0), "=l" (z1) : "r" (a), "r" (b)); \
   } while(0);

#endif

namespace Botan {

/*************************************************
* Word Multiply                                  *
*************************************************/
inline word word_mul(word a, word b, word* c)
   {
   word z0 = 0, z1 = 0;

#if defined(BOTAN_WORD_MUL)
   BOTAN_WORD_MUL(a, b, z1, z0);
#else
   bigint_wordmul(a, b, &z1, &z0);
#endif

   z1 += *c; if(z1 < *c) z0++;

   *c = z0;
   return z1;
   }

/*************************************************
* Word Multiply/Add                              *
*************************************************/
inline void word_madd(word a, word b, word c, word d,
                      word* out_low, word* out_high)
   {
   word z0 = 0, z1 = 0;

#if defined(BOTAN_WORD_MUL)
   BOTAN_WORD_MUL(a, b, z1, z0);
#else
   bigint_wordmul(a, b, &z1, &z0);
#endif

   z1 += c; if(z1 < c) z0++;
   z1 += d; if(z1 < d) z0++;

   *out_low = z1;
   *out_high = z0;
   }

/*************************************************
* Multiply-Add Accumulator                       *
*************************************************/
inline void word3_muladd(word* w2, word* w1, word* w0, word x, word y)
   {
   word t1 = 0;
   word_madd(x, y, *w0, 0, w0, &t1);
   *w1 += t1;
   *w2 += (*w1 < t1) ? 1 : 0;
   }

/*************************************************
* Multiply-Add Accumulator                       *
*************************************************/
inline void word3_muladd_2(word* w2, word* w1, word* w0, word x, word y)
   {
   word t0 = 0, t1 = 0;
   word_madd(x, y, 0, 0, &t0, &t1);

   *w0 += t0;
   *w1 += t1 + ((*w0 < t0) ? 1 : 0);
   *w2 += (*w1 < t1) ? 1 : 0;

   *w0 += t0;
   *w1 += t1 + ((*w0 < t0) ? 1 : 0);
   *w2 += (*w1 < t1) ? 1 : 0;
   }

}

#endif
